"""Cross-python stdlib shims."""

import collections.abc as _c  # noqa: WPS111, WPS301
import os
import sys
from contextlib import contextmanager
from pathlib import Path


if sys.version_info >= (3, 11):
    from contextlib import chdir as chdir_cm

    from tomllib import loads as load_toml_from_string
else:
    from tomli import loads as load_toml_from_string

    @contextmanager  # type: ignore[no-redef]
    def chdir_cm(path: 'os.PathLike[str]') -> _c.Iterator[None]:
        """Temporarily change the current directory, recovering on exit."""
        original_wd = Path.cwd()
        os.chdir(path)
        try:  # noqa: WPS505
            yield
        finally:
            os.chdir(original_wd)


__all__ = (  # noqa: WPS410
    'chdir_cm',
    'load_toml_from_string',
)



import pickle

from yarl import URL

# serialize


def test_pickle() -> None:
    u1 = URL("picklepickle")
    hash(u1)
    v = pickle.dumps(u1)
    u2 = pickle.loads(v)
    assert u1._cache
    assert not u2._cache
    assert hash(u1) == hash(u2)


def test_default_style_state() -> None:
    u = object.__new__(URL)
    val = ("set_state", "set_state", "set_state", "set_state", "set_state")
    u.__setstate__((None, {"_val": val}))
    assert u._val == val
    assert hash(u) != 1


def test_empty_url_is_not_cached() -> None:
    u = URL.__new__(URL)
    val = ("set_state", "set_state", "set_state", "set_state", "set_state")
    u.__setstate__((None, {"_val": val}))
    assert u._val == val
    assert hash(u) != 1


def test_pickle_does_not_pollute_cache() -> None:
    """Verify the unpickling does not pollute the cache.

    Since unpickle will call URL.__new__ with default
    args, we need to make sure that default args never
    end up in the pre_encoded_url or encode_url cache.
    """
    u1 = URL.__new__(URL)
    u1._scheme = "this"
    u1._netloc = "never.appears.any.where.else.in.tests"
    u1._path = ""
    u1._query = ""
    u1._fragment = ""
    hash(u1)
    v = pickle.dumps(u1)
    u2: URL = pickle.loads(v)
    assert u1._cache
    assert hash(u1) == hash(u2)
    assert u2._scheme == "this"
    assert u2._netloc == "never.appears.any.where.else.in.tests"
    assert u2._path == ""
    assert u2._query == ""
    assert u2._fragment == ""
    # Verify unpickling did not the cache wrong scheme
    # for empty args.
    assert URL().scheme == ""
    assert URL("").scheme == ""



from typing import TYPE_CHECKING

import pytest

from yarl import URL

if TYPE_CHECKING:
    import pydantic
else:
    pydantic = pytest.importorskip("pydantic")


class TstModel(pydantic.BaseModel):
    url: URL


def test_dump() -> None:
    url = URL("https://example.com")
    m = TstModel(url=url)
    dct = m.model_dump()
    assert dct == {"url": str(url)}
    assert isinstance(dct["url"], str)


def test_validate_valid() -> None:
    url = URL("https://example.com")
    dct = {"url": str(url)}
    m = TstModel.model_validate(dct)
    assert m == TstModel(url=url)
    assert isinstance(m.url, URL)


def test_validate_invalid() -> None:
    dct = {"url": 123}
    with pytest.raises(pydantic.ValidationError, match="url"):
        TstModel.model_validate(dct)


def test_get_schema() -> None:
    schema = TstModel.model_json_schema()
    assert schema == {
        "properties": {"url": {"format": "uri", "title": "Url", "type": "string"}},
        "required": ["url"],
        "title": "TstModel",
        "type": "object",
    }


def test_json_roundtrip_json() -> None:
    url = URL("https://example.com")
    m = TstModel(url=url)
    js = m.model_dump_json()
    m2 = TstModel.model_validate_json(js)
    assert m == m2
    js2 = m2.model_dump_json()
    assert js == js2


def test_fake_cover() -> None:
    # The test exists only for getting ocverage for __get_pydantic_core_schema__,
    # otherwise a call of python code back from rust is not measured
    # by coverage tool

    URL.__get_pydantic_core_schema__(URL, pydantic.GetCoreSchemaHandler())



"""PEP 517 build backend for optionally pre-building Cython."""

from contextlib import suppress as _suppress

# Re-exporting PEP 517 hooks
# pylint: disable-next=unused-wildcard-import,wildcard-import
from setuptools.build_meta import *  # noqa: F403, WPS347

# Re-exporting PEP 517 hooks
from ._backend import (  # type: ignore[assignment]
    build_sdist,  # noqa: F401
    build_wheel,  # noqa: F401
    get_requires_for_build_wheel,  # noqa: F401
    prepare_metadata_for_build_wheel,  # noqa: F401
)


with _suppress(
    ImportError,
):  # Only succeeds w/ setuptools implementing PEP 660
    # Re-exporting PEP 660 hooks
    from ._backend import (  # type: ignore[assignment]
        build_editable,  # noqa: F401
        get_requires_for_build_editable,  # noqa: F401
        prepare_metadata_for_build_editable,  # noqa: F401
    )



# fmt: off

from __future__ import annotations

import os
import typing as _t  # noqa: WPS111
from contextlib import contextmanager
from pathlib import Path
from sys import version_info as _python_version_tuple

from expandvars import expandvars

from ._compat import load_toml_from_string
from ._transformers import (
    get_cli_kwargs_from_config,
    get_enabled_cli_flags_from_config,
)


if _t.TYPE_CHECKING:
    import collections.abc as _c  # noqa: WPS111, WPS301


class Config(_t.TypedDict):
    """Data structure for the TOML config."""

    env: dict[str, str]
    flags: dict[str, bool]
    kwargs: dict[str, str | dict[str, str]]
    src: list[str]


def get_local_cython_config() -> Config:
    """Grab optional build dependencies from pyproject.toml config.

    :returns: config section from ``pyproject.toml``
    :rtype: dict

    This basically reads entries from::

        [tool.local.cython]
        # Env vars provisioned during cythonize call
        src = ["src/**/*.pyx"]

        [tool.local.cython.env]
        # Env vars provisioned during cythonize call
        LDFLAGS = "-lssh"

        [tool.local.cython.flags]
        # This section can contain the following booleans:
        # * annotate — generate annotated HTML page for source files
        # * build — build extension modules using distutils
        # * inplace — build extension modules in place using distutils (implies -b)
        # * force — force recompilation
        # * quiet — be less verbose during compilation
        # * lenient — increase Python compat by ignoring some compile time errors
        # * keep-going — compile as much as possible, ignore compilation failures
        annotate = false
        build = false
        inplace = true
        force = true
        quiet = false
        lenient = false
        keep-going = false

        [tool.local.cython.kwargs]
        # This section can contain args that have values:
        # * exclude=PATTERN      exclude certain file patterns from the compilation
        # * parallel=N    run builds in N parallel jobs (default: calculated per system)
        exclude = "**.py"
        parallel = 12

        [tool.local.cython.kwargs.directives]
        # This section can contain compiler directives
        # NAME = "VALUE"

        [tool.local.cython.kwargs.compile-time-env]
        # This section can contain compile time env vars
        # NAME = "VALUE"

        [tool.local.cython.kwargs.options]
        # This section can contain cythonize options
        # NAME = "VALUE"
    """
    config_toml_txt = (Path.cwd().resolve() / 'pyproject.toml').read_text()
    config_mapping = load_toml_from_string(config_toml_txt)
    return config_mapping['tool']['local']['cython']  # type: ignore[no-any-return]


def get_local_cythonize_config() -> Config:
    """Grab optional build dependencies from pyproject.toml config.

    :returns: config section from ``pyproject.toml``
    :rtype: dict

    This basically reads entries from::

        [tool.local.cythonize]
        # Env vars provisioned during cythonize call
        src = ["src/**/*.pyx"]

        [tool.local.cythonize.env]
        # Env vars provisioned during cythonize call
        LDFLAGS = "-lssh"

        [tool.local.cythonize.flags]
        # This section can contain the following booleans:
        # * annotate — generate annotated HTML page for source files
        # * build — build extension modules using distutils
        # * inplace — build extension modules in place using distutils (implies -b)
        # * force — force recompilation
        # * quiet — be less verbose during compilation
        # * lenient — increase Python compat by ignoring some compile time errors
        # * keep-going — compile as much as possible, ignore compilation failures
        annotate = false
        build = false
        inplace = true
        force = true
        quiet = false
        lenient = false
        keep-going = false

        [tool.local.cythonize.kwargs]
        # This section can contain args that have values:
        # * exclude=PATTERN      exclude certain file patterns from the compilation
        # * parallel=N    run builds in N parallel jobs (default: calculated per system)
        exclude = "**.py"
        parallel = 12

        [tool.local.cythonize.kwargs.directives]
        # This section can contain compiler directives
        # NAME = "VALUE"

        [tool.local.cythonize.kwargs.compile-time-env]
        # This section can contain compile time env vars
        # NAME = "VALUE"

        [tool.local.cythonize.kwargs.options]
        # This section can contain cythonize options
        # NAME = "VALUE"
    """
    config_toml_txt = (Path.cwd().resolve() / 'pyproject.toml').read_text()
    config_mapping = load_toml_from_string(config_toml_txt)
    return config_mapping['tool']['local']['cythonize']  # type: ignore[no-any-return]


def _configure_cython_line_tracing(
    config_kwargs: dict[str, str | dict[str, str]],
    *,
    cython_line_tracing_requested: bool,
) -> None:
    """Configure Cython line tracing directives if requested."""
    # If line tracing is requested, add it to the directives
    if cython_line_tracing_requested:
        directives = config_kwargs.setdefault('directive', {})
        assert isinstance(directives, dict)  # noqa: S101  # typing
        directives['linetrace'] = 'True'
        directives['profile'] = 'True'


def make_cythonize_cli_args_from_config(
    config: Config,
    *,
    cython_line_tracing_requested: bool = False,
) -> list[str]:
    """Compose ``cythonize`` CLI args from config."""
    py_ver_arg = f'-{_python_version_tuple.major!s}'

    cli_flags = get_enabled_cli_flags_from_config(config['flags'])
    config_kwargs = config['kwargs']

    _configure_cython_line_tracing(
        config_kwargs,
        cython_line_tracing_requested=cython_line_tracing_requested,
    )

    cli_kwargs = get_cli_kwargs_from_config(config_kwargs)

    return cli_flags + [py_ver_arg] + cli_kwargs + ['--'] + config['src']


@contextmanager
def patched_env(
    env: dict[str, str],
    *,
    cython_line_tracing_requested: bool,
    original_source_directory: Path | None = None,
    temporary_build_directory: Path | None = None,
) -> _c.Iterator[None]:
    """Temporary set given env vars.

    :param env: tmp env vars to set
    :type env: dict

    :yields: None
    """
    orig_env = os.environ.copy()
    expanded_env = {name: expandvars(var_val) for name, var_val in env.items()}  # type: ignore[no-untyped-call]
    os.environ.update(expanded_env)

    os.environ['CFLAGS'] = ' '.join((
        # First, low priority hardcoded value from the `pyproject.toml` config:
        expanded_env.get('CFLAGS', ''),
        # Next, add dynamically computed compiler flags:
        *(
            # Debug mode:
            (
                # Compiler-specific settings:
                '-g3',  # debug symbols w/ extra details
                '-Og',  # optimize for debug experience, better than -O0
                '-UNDEBUG',  # enable assertions
                # Coverage-related:
                '--coverage',
                # '-fkeep-inline-functions',  # clang seems to not support this
                # '-fkeep-static-functions',  # clang seems to not support this
                # '-fprofile-abs-path',  # clang seems to not support this
                # Cython-specific settings:
                '-DCYTHON_TRACE=1',
                '-DCYTHON_TRACE_NOGIL=1',
            )
            if cython_line_tracing_requested
            # Release mode:
            else (
                '-g0',  # no debug symbols
                '-Ofast',  # maximum optimization
                '-DNDEBUG',  # disable assertions
            )
        ),
        *(
            # In-tree mode:
            ()
            if temporary_build_directory is None
            # Temporary build directory mode:
            else (
                '-ffile-prefix-map='
                f'{temporary_build_directory!s}={original_source_directory!s}',
            )
        ),
        # Finally, append the user-set env var, ensuring its top priority:
        orig_env.get('CFLAGS', ''),
        # Last thing, strip spaces caused by empty leading/trailing flags:
    )).strip()

    os.environ['LDFLAGS'] = ' '.join((
        # First, low priority hardcoded value from the `pyproject.toml` config:
        expanded_env.get('LDFLAGS', ''),
        # Next, add dynamically computed linker flags:
        *(
            # Debug mode:
            (
                # Coverage-related:
                '--coverage',
            )
            if cython_line_tracing_requested
            # Release mode:
            else (
                '-s',  # remove all symbol table and relocation information
            )
        ),
        # Finally, append the user-set env var, ensuring its top priority:
        orig_env.get('LDFLAGS', ''),
        # Last thing, strip spaces caused by empty leading/trailing flags:
    )).strip()

    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(orig_env)



"""Data conversion helpers for the in-tree PEP 517 build backend."""

from __future__ import annotations

import typing as _t  # noqa: WPS111
from itertools import chain
from re import sub as _substitute_with_regexp


if _t.TYPE_CHECKING:
    import collections.abc as _c  # noqa: WPS111, WPS301


def _emit_opt_pairs(
    opt_pair: tuple[str, dict[str, str] | str],
) -> _c.Iterator[str]:
    flag, flag_value = opt_pair
    flag_opt = f'--{flag!s}'
    if isinstance(flag_value, dict):
        sub_pairs: _c.Iterable[tuple[str, ...]] = flag_value.items()
    else:
        sub_pairs = ((flag_value,),)

    for pair in sub_pairs:
        yield '='.join(map(str, (flag_opt, *pair)))


def get_cli_kwargs_from_config(
    kwargs_map: dict[str, str | dict[str, str]],
) -> list[str]:
    """Make a list of options with values from config."""
    return list(chain.from_iterable(map(_emit_opt_pairs, kwargs_map.items())))


def get_enabled_cli_flags_from_config(
    flags_map: _c.Mapping[str, bool],
) -> list[str]:
    """Make a list of enabled boolean flags from config."""
    return [
        f'--{flag}' for flag, is_enabled in flags_map.items() if is_enabled
    ]


def sanitize_rst_roles(rst_source_text: str) -> str:  # noqa: WPS210
    """Replace RST roles with inline highlighting."""
    pep_role_regex = r"""(?x)
        :pep:`(?P<pep_number>\d+)`
    """
    pep_substitution_pattern = (
        r'`PEP \g<pep_number> <https://peps.python.org/pep-\g<pep_number>>`__'
    )

    user_role_regex = r"""(?x)
        :user:`(?P<github_username>[^`]+)(?:\s+(.*))?`
    """
    user_substitution_pattern = (
        r'`@\g<github_username> '
        r'<https://github.com/sponsors/\g<github_username>>`__'
    )

    issue_role_regex = r"""(?x)
        :issue:`(?P<issue_number>[^`]+)(?:\s+(.*))?`
    """
    issue_substitution_pattern = (
        r'`#\g<issue_number> '
        r'<https://github.com/aio-libs/yarl/issues/\g<issue_number>>`__'
    )

    pr_role_regex = r"""(?x)
        :pr:`(?P<pr_number>[^`]+)(?:\s+(.*))?`
    """
    pr_substitution_pattern = (
        r'`PR #\g<pr_number> '
        r'<https://github.com/aio-libs/yarl/pull/\g<pr_number>>`__'
    )

    commit_role_regex = r"""(?x)
        :commit:`(?P<commit_sha>[^`]+)(?:\s+(.*))?`
    """
    commit_substitution_pattern = (
        r'`\g<commit_sha> '
        r'<https://github.com/aio-libs/yarl/commit/\g<commit_sha>>`__'
    )

    gh_role_regex = r"""(?x)
        :gh:`(?P<gh_slug>[^`<]+)(?:\s+([^`]*))?`
    """
    gh_substitution_pattern = r'GitHub: ``\g<gh_slug>``'

    meth_role_regex = r"""(?x)
        (?::py)?:meth:`~?(?P<rendered_text>[^`<]+)(?:\s+([^`]*))?`
    """
    meth_substitution_pattern = r'``\g<rendered_text>()``'

    role_regex = r"""(?x)
        (?::\w+)?:\w+:`(?P<rendered_text>[^`<]+)(?:\s+([^`]*))?`
    """
    substitution_pattern = r'``\g<rendered_text>``'

    project_substitution_regex = r'\|project\|'
    project_substitution_pattern = 'yarl'

    substitutions = (
        (pep_role_regex, pep_substitution_pattern),
        (user_role_regex, user_substitution_pattern),
        (issue_role_regex, issue_substitution_pattern),
        (pr_role_regex, pr_substitution_pattern),
        (commit_role_regex, commit_substitution_pattern),
        (gh_role_regex, gh_substitution_pattern),
        (meth_role_regex, meth_substitution_pattern),
        (role_regex, substitution_pattern),
        (project_substitution_regex, project_substitution_pattern),
    )

    rst_source_normalized_text = rst_source_text
    for regex, substitution in substitutions:
        rst_source_normalized_text = _substitute_with_regexp(
            regex,
            substitution,
            rst_source_normalized_text,
        )

    return rst_source_normalized_text
